/*===========================================================================
   Pixelize 1.0 - by Santiago Romero aka NoP / Compiler   (c) 2001.
   
      Pixelize is a simple program I wrote to "pixelize" black & white
      and colour images in order to allow people doing phisical mosaics
      of real images... it can print on the screen the used colours in
      order to build the mosaic externally...

      The most spectacular example of this kind of mosaics is the
      "Mona Lisa" picture done using toasts (colours = different cooking
      time from cool to burnt :) on a European city.
  
      
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 Copyright (c) 2000 Santiago Romero Iglesias
 Email: sromero@escomposlinux.org


 Usage example:

  ./pixelize test.bmp 8 8 16 > output.txt
  
===========================================================================*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <allegro.h>

/* global variables used */
BITMAP *img;
PALETTE palette;
int window_width=800, window_height=600;
int img_w, img_h;
int square_w, square_h;
int num_squares_x, num_squares_y;
int num_colors;
#define NUM_ASCII 16
char ascii_colour[NUM_ASCII] = 
  { ' ', '.', ',', '.', ';', '^', 'o', '+', 'x',
    '*', '#', 'Q', '@', 'O', 'X', 'W' };
                           
/* some useful macros */
#define SCREEN_WIDTH  800 
#define SCREEN_HEIGHT 600
#define RED(pal,col)   (pal[(col)].r)  
#define GREEN(pal,col) (pal[(col)].g)
#define BLUE(pal,col)  (pal[(col)].b)
#define CopyScreen(org,dst) blit((org),(dst),0,0,0,0,(org)->w,(org)->h)

/* some function declarations here */
void InitAll( void );
void LoadAll( char * );
void ExitAll( void );
int pixelize( BITMAP *image, int w, int h, 
              int square_width, int square_height, int numcolors );
int displayImagetoText( BITMAP *image, int w, int h,
                        int square_width, int square_height, int numcolors );
void doPixelize( int x0, int y0, int square_w,
                 int square_h, int numcolors, BITMAP *image );
int NearestColor( int r, int g, int b, PALETTE paleta, int colors );
int reduceColours( BITMAP *image, int squarew, int squareh, 
                   int numcolors, PALETTE pal );
   


/*---------------------------------------------------------------------------
 Main function:
  - Read command line arguments.
  - Open image file (.BMP or .PCX) and load it.
  - Do the pixelization process.
  - Write the results on the screen (can be redirected to file).
---------------------------------------------------------------------------*/
int main( int argc, char *argv[] )
{
   
    printf("Pixelize 1.0 - by Santiago Romero aka NoP / Compiler   (c) 2001.\n\n");
    // check arguments
    if( argc < 5 )
    {
      printf("Pixelize Usage: \n");
      printf("\t./pixelize image.bmp square_width square_height colours\n");
      exit(-1);
    }
    
    // Init Allegro stuff, load the bitmap from disk and setup variables
    InitAll();
    LoadAll(argv[1]);
    
    img_w = img->w;
    img_h = img->h;
    square_w = atoi(argv[2]);
    square_h = atoi(argv[3]);
    num_colors = atoi( argv[4]);
    num_squares_x = img_w / square_w;
    num_squares_y = img_h / square_h;

    printf("Image       : %s\n", argv[1]);
    printf("Image size  : %d x %d\n", img_w, img_h );
    printf("Square size : %d x %d\n", square_w, square_h );
    printf("Squares     : %d x %d\n", num_squares_x, num_squares_y );
    printf("Colours     : %d\n\n", num_colors );
    
    // do the pixelization process
    readkey(); 
    pixelize( img, img_w, img_h, square_w, square_h, num_colors );
    CopyScreen( img, screen );
    readkey();
    
    reduceColours( img, square_w, square_h, num_colors, palette );
    CopyScreen( img, screen );
    readkey();
    
    // display the text output on the screen
    displayImagetoText( img, img_w, img_h, square_w, square_h, num_colors );
    ExitAll();
}
END_OF_MAIN();



/*---------------------------------------------------------------------------
 Pixelize function. It does all the work. 
---------------------------------------------------------------------------*/
int pixelize( BITMAP *image, int w, int h,
              int square_width, int square_height, int numcolors )
{
   int x, y;

   // divide the image in squares and apply doPixelize on them:
   for( y=0; y<h; y+=square_height )
      for( x=0; x<w; x+=square_width )
        doPixelize( x, y, square_width, square_height, numcolors, image );
    
}



/*---------------------------------------------------------------------------
 Main rendering function: 
   - Read all the pixels on a square and:
      r += pixel_r;     g += pixel_g;        b += pixel_g;
   - Find the most used (r,g,b) resultant colour.
   - Redraw the square using that colour.
---------------------------------------------------------------------------*/
void doPixelize( int x0, int y0, int square_w, int square_h,
                 int numcolors, BITMAP *image )
{
   int x, y;
   int media = 0, color;
   int media_r = 0, media_g = 0, media_b = 0;
  
   // calculate the media_r, g and b for a given block
   for( y=0; y<square_h; y++ )
   {
      for( x=0; x<square_w; x++ )
      {
         color = getpixel( image, x0+x, y0+y );
         media_r += RED(palette, color);
         media_g += GREEN(palette, color);
         media_b += BLUE(palette, color);
  
      }
   }
   
   media_r = media_r / (square_w * square_h);
   media_g = media_g / (square_w * square_h);
   media_b = media_b / (square_w * square_h);
   media = NearestColor( media_r, media_g, media_b, palette, 255 );
   
   // repaint all the square using the media
   rectfill( image, x0, y0, x0+square_w, y0+square_h, media );
   
}


/*---------------------------------------------------------------------------
 Reduces the number of colours used on the image to a given number.
 To do that it does an histogram of the most used colours, sorting the
 palette array to use that colours and forcing NearestColor to search
 just in the first N entries.

 Another way to do this is prepare a special palette with simple colours
 in the first 16 slots, and adding more colours for the rest of the palette.
 This should then distinguish between gray and colour images and use two
 different palettes.
---------------------------------------------------------------------------*/
int reduceColours( BITMAP *image, int squarew, int squareh, 
                   int numcolors, PALETTE pal )
{
   int x, y, block, i, j, temp, temp2, media;
   struct color
   {
      int palindex, counted;
   };
   struct color histogram[256];
   PALETTE newpal;

   // clear the color counter/histogram
   for( x=0; x<256; x++ )
   {
      histogram[x].palindex = x;
      histogram[x].counted = 0;
   }
   
   // divide the image in squares and count the apparition frequency:
   for( y=0; y<img->h; y+=squareh )
      for( x=0; x<img->w; x+=squarew )
   {
      block = getpixel( img, x, y );
      histogram[block].counted++;
   }

   // sort the colours (yes, old bubble sorting algorithm).
   for( i=1; i<256; i++ )
   {
        temp = histogram[i].counted;
        temp2 = histogram[i].palindex;
        j = i-1;

        while( temp>histogram[j].counted && j>=0 )
        {
            histogram[j+1].counted = histogram[j].counted;
            histogram[j+1].palindex = histogram[j].palindex;
            j = j-1;
        }
        histogram[j+1].counted = temp;
        histogram[j+1].palindex = temp2;
   }

   // Create newpal according to new colours:
   for( x=0; x<256; x++ )
   {
      newpal[x].r = RED(palette,histogram[x].palindex);
      newpal[x].g = GREEN(palette,histogram[x].palindex);
      newpal[x].b = BLUE(palette,histogram[x].palindex);
   }
          
   // repaint all the square using the most used colours
   for( y=0; y<image->h; y+=squareh )
   {
     for( x=0; x<image->w; x+=squarew )
     {
        block = getpixel( screen, x, y );
        media = NearestColor( RED(palette,block),
                              GREEN(palette,block), 
                              BLUE(palette,block),
                              newpal, numcolors );
        rectfill( image, x, y, x+squarew, y+squareh, media );
     }
   }
   set_palette( newpal );
 
}


/*---------------------------------------------------------------------------
 Show the image as numbers... this reads the image and displays it as
 a block of numbers ready to print.
---------------------------------------------------------------------------*/
int displayImagetoText( BITMAP *image, int w, int h,
              int square_width, int square_height, 
              int numcolors )
{
   int x, y, z, color;
   PALETTE newpal;

   get_palette( newpal );
   
   // format the data using --- and | .
   
   for( z=0; z<(3*w/square_width)+1; z++ ) printf("-");
   printf("---");

   for( y=0; y<h-square_height; y+=square_height )
   {
      printf("\n|");
      
      for( x=0; x<w-square_width; x+=square_width )
      {
         //color = getpixel( image, x, y ) / (256/numcolors);
         color = getpixel( image, x, y );
         if( numcolors < 99 )
         {
             if( color != 0 ) printf("%02d|", color );
             else             printf("  |");
         }
      }  
   }
      
   printf("\n");
   for( z=0; z<(3*w/square_width)+1; z++ )
        printf("-");
   printf("---\n");
   
   if( numcolors <= NUM_ASCII )
   {
   for( y=0; y<h; y+=square_height )
   {
         printf("\n");
         for( x=0; x<w; x+=square_width )
         {
            color = getpixel( image, x, y );
            printf( "%c", ascii_colour[color] );
         }  
      }
   }
   
   printf("\nPalette used:\n");
   for( x=0; x<numcolors; x++ )
     printf("%3d: (%d,%d,%d)\n", x, 
            RED(newpal,x), GREEN(newpal,x), BLUE(newpal,x));

}



/*---------------------------------------------------------------------------
 Allegro Initialization function (inits the graphical system).
---------------------------------------------------------------------------*/
void InitAll( void )
{
   int i;
   
   srand(time(NULL));
   allegro_init();
   install_keyboard();

   set_color_depth(8);
   if (set_gfx_mode(GFX_AUTODETECT, window_width, window_height, 0, 0) != 0)
   {
      set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
      allegro_message("Error setting graphics mode\n%s\n", allegro_error);
      ExitAll();
   }
   
   clear(screen);
}





/*---------------------------------------------------------------------------
 Load function: loads the image from file.
---------------------------------------------------------------------------*/
void LoadAll( char *filename )
{
   img = load_bitmap( filename, palette );
   set_palette( palette );
   CopyScreen( img, screen );
}


/*---------------------------------------------------------------------------
 Exit Function. It exits Pixelize ... :)
---------------------------------------------------------------------------*/
void ExitAll( void )
{
   if( img != NULL ) destroy_bitmap(img);
   set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
   allegro_exit();
   printf("---");
   exit(0);
}


/*---------------------------------------------------------------------------
 int NearestColor( r, g, b, paleta );
 Looks for the nearest (more similar) colour on a palette, giving it
 the (r,g,b) values of the colour to search. It returns -1 on error (it's
 almost impossible to fail).
---------------------------------------------------------------------------*/

int NearestColor( int r, int g, int b, PALETTE paleta, int colors )
{

  int result = -1;       // color mnimo hasta ahora
  int minimo = 200000;   // distancia mnima
  int rd, gd, bd;        // componentes vector c2-c1
  int dist, i;           // distancia e ndice bucle
                                    
  // repeat for N colours:
  for( i=0 ; i<colors ; i++ )
  {
    // calculate vector components as r2-r1, etc
    rd = r-RED(paleta, i);
    gd = g-GREEN(paleta, i);
    bd = b-BLUE(paleta, i);
  
    // calculate colour distances:  
    dist = rd*rd + gd*gd + bd*bd;

    // check if this distance is the smaller till now
    if( dist <minimo )
    {
      // is the lower: annotate and continue.
      minimo = dist ;
      result = i;
    }
  }               
  return( result );
}

